%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main testing script for the QSwift Library
% IrisCores.com | Author: vojta.ters@iriscores.com
% Free to use. Requires MATLAB R2016a+ 64-bit and supported nVidia graphics card
% In case of failure,try installing MS Redistributables for Visual Studio 2019
% If it doesnt solve the problem, try installing new nVidia drivers from
% http://www.nvidia.co.uk
% 
% Function overview:
% 1, Matrix Matrix Multiply
% 2, Gaussian Elimination 
% 3, Matrix Inversion
% 4, Hauseholder Transformation
% 5, Gram-Schmidt Decomposition
% 6, Eigenvalue Computation
% 7, Jacobi Method - Iteration based Linear System Solver
% 8, SuperRelaxation - Iteration Based Linear System Solver
%
% All functions implemented work only in single precision mode.
% I do not guarantee any faster computation.
% There is however always plenty of space for optimizations.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

close all;
clear all;
format bank;
clc;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Add release Directory to Matlab Path
MexFilePath = ['..',filesep,'x64',filesep,'Release'];
addpath(MexFilePath);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% Get Display Resolution %%%%%%%
Displ = groot();
ScW = Displ.ScreenSize(3);
ScH = Displ.ScreenSize(4);
FiW = 1024;
FiH = 720;
Cls = perms([0,102,204])/255;
Cl_Brown      = 1;
Cl_Dark_Red   = 2;
Cl_Dark_Green = 3;
Cl_Violet     = 4;
Cl_Green      = 5;
Cl_Blue       = 6; 
Lw            = 1;  % LineWidth
Ms            = 13; % MarkerSize
Fs            = 17; % FontSize 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Display Border Limitations %
Ma            = 0.05;
Left          = Ma;
Bot           = Ma;

%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% %%%%%%% MATRIX MATRIX MULTIPLICATION %%%%%%% %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
N   = 20;
A   = randn(N);B = randn(N);
Man = QSwiftMatlab('IData1',single(A),'IData2',single(B),'Type','MatrixMatrixMul');
clear mex;
Ref = A*B;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Calculate Relative Difference %
Rel = ((Man - Ref) ./ Ref) * 100;
Rel_Max = max(Rel(:));
disp(['CUDA - Relative Difference [MM - Multiply]: ',num2str(Rel_Max),' %']);
figure('name','Matrix Matrix Multiplication','position',[ScW/2-FiW/2,ScH/2-FiH/2,FiW,FiH]);
mesh(Rel,'FaceAlpha',0.7);
title('Relative Difference','FontSize',Fs);
colormap morgenstemning;

%%%%%%%%%%%%%%%%%%%%%%%%%
% Update Figure Borders %
set(gca,'Position',[Left Bot 1 - 2*Ma 1 - 2*Ma]);

%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%% %%%%%%% JACOBI LINEAR SYSTEM SOLVER %%%%%%% %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
N = 10;                      % Matrix Dimension
A = rand(N) -0.5 + eye(N);   % Matrix of Equations
B = rand(N,1);               % Right Side
NumIter = 50;                % Maximum Amount of Iterations
Precision = 1e-7;            % Required Precision

A = single(A);
B = single(B);
X0 = zeros(N,1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% Reference Matlab Solver %%%%%%% 
Ref = JacobiSolver(A,B,X0,NumIter,Precision);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% Custom CUDA-Accelerated Solver %%%%%%
Man = QSwiftMatlab('Type','Jacobi','IData1',A,'IData2',B,'NumIter',NumIter,'Precision',1e-3);
clear mex;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Calculate Relative Difference %
Rel = ((Man - Ref) ./ Ref) * 100;
Rel_Max = max(Rel(:));
disp(['CUDA - Relative Difference [JACOBI]: ',num2str(Rel_Max),' %']);
figure('name','Jacobi - IterationSolver','position',[ScW/2-FiW/2,ScH/2-FiH/2,FiW,FiH]);
plot(Rel,'color',Cls(Cl_Violet,:),'linewidth',Lw,'MarkerSize',Ms);
title('Difference Matlab - QSwift','FontSize',Fs);
grid on;

%%%%%%%%%%%%%%%%%%%%%%%%%
% Update Figure Borders %
set(gca,'Position',[Left Bot 1 - 2*Ma 1 - 2*Ma]);

%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% %%%%%%% SUCCESSIVE OVER RELAXATION METHOD %%%%%%% %%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

N = 10;                 % Matrix Dimension
A = rand(N) + eye(N);   % Matrix of Equations
B = rand(N,1) - 0.5;    % Right Side
G0 = zeros(N,1);        % Initial Guess
NumIter = 50;           % Maximum Amount of Iterations
Omega = 1.3;            % Super-Relaxation Coefficient
Precision = 1e-5;       % Required Precision

A = single(A);
B = single(B);
G0 = single(G0);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% Reference Matlab Solver %%%%%%%
Ref = SuccessiveOverRelax(A,B,G0,NumIter,Omega,Precision); 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% Custom CUDA-Accelerated Solver %%%%%%%
Man = QSwiftMatlab('Type','Sor','IData1', A ,'IData2', B ,'NumIter',NumIter,'Omega',Omega,'Precision',Precision);
clear mex;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Calculate Relative Difference %
Rel = ((Man - Ref) ./ Ref) * 100;
Rel_Max = max(Rel(:));
disp(['CUDA - Relative Difference [SOR]: ',num2str(Rel_Max),' %']);
figure('name','Successive OverRelaxation - Iteration Solver','position',[ScW/2-FiW/2,ScH/2-FiH/2,FiW,FiH]);
plot(Rel,'color',Cls(Cl_Blue,:),'LineWidth',Lw);
title('Difference Matlab - QSwift');
grid on;

%%%%%%%%%%%%%%%%%%%%%%%%%
% Update Figure Borders %
set(gca,'Position',[Left Bot 1 - 2*Ma 1 - 2*Ma]);


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%% %%%%%%% GAUSSIAN ELIMINATION %%%%%%% %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
N = 20;                 % Matrix Dimension
A = 1 + rand(N) - 0.5;  % Random Matrix
A = single(A);          % Convert to Single-Precision
Prec = 1e-5;            % Just for Visualization

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% Reference Matlab Solver %%%%%%% 
Man = GaussElimForw(A);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% Custom CUDA-Accelerated Solver %%%%%%
Ref = QSwiftMatlab('IData1',A,'Type','GaussElim');
clear mex;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Calculate Relative Difference %
Rel = ((Man - Ref) ./ Ref) * 100;
Rel(abs(Ref) < Prec) = 0; % Mask Zeros / Nan / Inf

Rel_Max = max(Rel(:));
disp(['CUDA - Relative Difference [GAUSSELIM]: ',num2str(Rel_Max),' %']);
figure('name','Gaussian Elimination','position',[ScW/2-FiW/2,ScH/2-FiH/2,FiW,FiH]);
mesh(Rel,'FaceAlpha',0.7);
title('GaussElim Difference');
colormap hsv;

%%%%%%%%%%%%%%%%%%%%%%%%%
% Update Figure Borders %
set(gca,'Position',[Left Bot 1 - 2*Ma 1 - 2*Ma]);


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% %%%%%%% MATRIX INVERSION %%%%%%% %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
N = 10;             % Matrix Dimension
A = rand(N) - 0.5;  % Random Matrix
A = single(A);      % Convert to Single Precision

Ref = inv(A);
Man = QSwiftMatlab('IData1',single(A),'Type','MatrixInverse');
clear mex;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Calculate Relative Difference %
ShouldBeEye = Man*A; % inv(A) * A = E
Rel = ((Man - Ref) ./ Ref) * 100;
Rel_Max = max(Rel(:));
disp(['CUDA - Relative Difference [MM - Inv]: ',num2str(Rel_Max),' %']);
figure('name','Matrix Inversion','position',[ScW/2-FiW/2,ScH/2-FiH/2,FiW,FiH]);
surf(ShouldBeEye,'FaceAlpha',0.7);
title('EYE = inv(A)xA','FontSize',Fs);

%%%%%%%%%%%%%%%%%%%%%%%%%
% Update Figure Borders %
set(gca,'Position',[Left Bot 1 - 2*Ma 1 - 2*Ma]);

%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% %%%%%%% GRAM SCHMIDT DECOMPOSITION %%%%%%% %%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
N = 1000;                % Matrix Dimension
A = ones(N) + randn(N); % Input Matrix
Prec = 1e-3; 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% Custom CUDA-Accelerated Solver %%%%%%
[Q,R] = QSwiftMatlab('IData1',single(A),'Type','GramSchmidt');
clear mex;

ECC = Q * R; % Should form Original Matrix
D = ECC - A; % Should be Zeros
MM = max(D(:));
if MM > Prec
    warning('Gram Schmidt Decomposition Tolerance out of bounds!');
end

figure('name','GramSchmidt - Q','position',[ScW/2-FiW/2,ScH/2-FiH/2,FiW,FiH]);
imshow(Q);
title('GramSchmidt - Q Matrix','FontSize',Fs);
set(gca,'Position',[Left Bot 1 - 2*Ma 1 - 2*Ma]);
colormap morgenstemning;

figure('name','GramSchmidt - R','position',[ScW/2-FiW/2,ScH/2-FiH/2,FiW,FiH]);
imshow(R);
title('GramSchmidt - R Matrix','FontSize',Fs);
set(gca,'Position',[Left Bot 1 - 2*Ma 1 - 2*Ma]);
colormap morgenstemning;



%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% %%%%%%% HAUSEHOLDER TRANSFORMATION %%%%%%% %%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
N = 1000;        % Matrix Dimension
A = randn(N);   % Random Matrix
A = single(A);  % Convert to Single Precision
Prec = 1e-4;    % Just for Visualization

%%%%%%%%%%%%%%%%%%%%%%%%%
% Make Symmetric Matrix %
% Symmetric Matrices are only 3-Diagonal
for i = 1:N
  for j = 1:N
    if(i < j)
      A(j,i) = A(i,j);
    end
  end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% Custom CUDA-Accelerated Solver %%%%%%
Man = QSwiftMatlab('IData1',A,'Type','HauseHolder');
clear mex;

%%%%%%%%%%%%%%%%%%%%%%%
% Calculate Reference %
tic
Ref = HauseholderTriang(A);
toc

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Calculate Relative Difference %
Rel = ((Man - Ref) ./ Ref) * 100;
Rel(abs(Ref) < Prec) = 0;
Rel_Max = max(Rel(:));
disp(['CUDA - Relative Difference [Tri - diagonalization]: ',num2str(Rel_Max),' %']);
figure('name','HauseHolder TriDiag - Matlkab ','position',[ScW/2-FiW/2,ScH/2-FiH/2,FiW,FiH]);
RefLog = log( 1 + abs(Ref));
imagesc(RefLog);
title('Matlab');
colormap morgenstemning;
set(gca,'Position',[Left Bot 1 - 2*Ma 1 - 2*Ma]);


figure('name','HauseHolder TriDiag - QSwift ','position',[ScW/2-FiW/2,ScH/2-FiH/2,FiW,FiH]);
ManLog = log( 1 + abs(Man));
imagesc(ManLog);
title('QSwift');
colormap morgenstemning;
set(gca,'Position',[Left Bot 1 - 2*Ma 1 - 2*Ma]);


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%% %%%%%%% EIGENVALUES COMPUTATION %%%%%%% %%%%%%%
N = 256;         % Matrix Dimension
A = randn(N);    % Random Matrix
A = single(A);   % Convert to Single Precision
NumIter = 50;    % Maximum amount of Iterations 

% MATLAB REFERENCE
AA = A;tic;
for i = 1:NumIter
    [Q,R] = qr(AA);
    AA = R*Q;
end

EiMat = diag(AA); % GET DIAGONAL ELEMENTS FROM MATLAB IMPL.
MatTime = toc;disp(['Matlab Elapsed Time: ',num2str(MatTime*1000),' ms']);
EiCuda = QSwiftMatlab('IData1',A,'Precision',1e-3,'NumIter',NumIter,'Type','Eigenvalues');
clear mex;

EIM = sort(EiMat);
EIC = sort(EiCuda');
figure('name','Eigenvalues Comparison','position',[ScW/2-FiW/2,ScH/2-FiH/2,FiW,FiH]);
subplot(2,1,1);
plot(EIM,'.','color',Cls(Cl_Green,:),'MarkerSize',Ms);
hold on;grid on;
plot(EIC,'-','color',Cls(Cl_Dark_Red,:));
legend('Matlab','QSwift');
subplot(2,1,2);
plot(abs(EIM-EIC),'color',Cls(Cl_Brown,:));
grid on;legend('Matlab - QSwift Difference');














