#pragma once

// ---------------------------
// ------ Cuda Includes ------
#include "cufft.h"
#include <cuda_runtime_api.h>

// --------------------------
// ------ C++ Includes ------
#include <algorithm>
#include <vector>
#include <string>

#define DLLEXPORT __declspec(dllexport)
#define DLLIMPORT __declspec(dllimport)
#define __FILENAME__ (strrchr(__FILE__, '/') ? strrchr(__FILE__, '/') + 1 : __FILE__)

#define CHECK_CUFFT(err)  (Handle_CuFFT(  err, __FILENAME__, __LINE__ ))
#define CHECK_CUDA(err)   (Handle_Error(  err, __FILENAME__, __LINE__ ))
#define CHECK_CUBLAS(err) (Handle_Cublas( err, __FILENAME__, __LINE__ ))

// -----------------------------------
// ------- Constant Definitions ------
static const size_t KILOPOW2 = 1024;
static const size_t MEGAPOW2 = 1048576;
static const size_t GIGAPOW2 = 1073741824;
#define FL_ZERO 0.0f
#define FL_ONE 1.0f
#define DBL_ZERO 0.0
#define DBL_ONE 1.0

namespace QNSwift {

	// ---------------------------------
	// ------- Shared Log Buffers ------
	static std::vector<std::string>	ErrorBuffer;
	static std::vector<std::string> MessageBuffer;

	/**************************************************************
	 * @brief: Defines the Available Computing Modes of the Library 
	 **************************************************************/
	enum class ComputingType {
		QS_METHOD_UNKNOWN		= 0,
		QS_METHOD_GAUSSELIM		= 1,
		QS_METHOD_INVERSION		= 2,
		QS_METHOD_MATRIXMUL		= 3,
		QS_METHOD_GRAM_SCHMIDT	= 4,
		QS_METHOD_HAUSEHOLDER	= 5,
		QS_METHOD_EIGENVALUES	= 6,
		QS_METHOD_JACOBIITER	= 7,
		QS_METHOD_SOR			= 8
	};

	/*****************************************************************************
	 * @brief: Defines Data ordering mechanism for 1D arrays storing 2D Structures
	 * COLUMN MAJOR linearly indexes columns while ROW MAJOR linearly indexes rows
	 *****************************************************************************/
	enum class ArrayFormat {QS_COLUMN_MAJOR,QS_ROW_MAJOR};
	
	/***********************************************************
	 *@brief: Defines the Array Complexity to be Real or Complex
	 **********************************************************/
	enum class ArrayComplexity { QS_REAL,QS_COMPLEX};

	/************************************************************
	 * @brief: Defines the Array Precision to be Single or Double
	 ************************************************************/
	enum class ArrayPrecision {QS_SINGLE_PRECISION,QS_DOUBLE_PRECISION};

	enum ParameterType {
		QS_GREEK_ALPHA		= 0,
		QS_GREEK_BETA		= 1,
		QS_GREEK_GAMMA		= 2,
		QS_GREEK_DELTA		= 3,
		QS_GREEK_EPSILON	= 4,
		QS_GREEK_LAMBDA		= 5,
		QS_GREEK_THETA		= 6,
		QS_GREEK_PHI		= 7,
		QS_GREEK_OMEGA		= 8,
		QS_GREEK_TAU		= 9,
		QS_GREEK_KAPPA		= 10,
		QS_NUM_ITER			= 11,
		QS_PRECISION		= 12

	};

	/*************************************
	 *@brief: Basically the same as "dim3" 
	 ************************************/
	struct Dimensions {

		Dimensions() {
			X = 0;
			Y = 0;
			Z = 0;
		}

		Dimensions(unsigned int  INx, unsigned int  INy, unsigned int  INz) {
			X = INx; Y = INy; Z = INz;
		}

		unsigned int X;
		unsigned int  Y;
		unsigned int  Z;
	};

	struct LaunchParams {

		LaunchParams(dim3 IBlocks,dim3 IThreads) {
			Blocks = IBlocks;
			Threads = IThreads;
			SharedMem = 0;
			Stream = cudaStreamDefault;
		}

		LaunchParams(dim3 IBlocks,dim3 IThreads,size_t ISharedMem) {
			Blocks = IBlocks;
			Threads = IThreads;
			SharedMem = ISharedMem;
			Stream = cudaStreamDefault;
		}

		LaunchParams(dim3 IBlocks,dim3 IThreads,size_t ISharedMem,cudaStream_t IStream) {
			Blocks = IBlocks;
			Threads = IThreads;
			SharedMem = ISharedMem;
			Stream = IStream;
		}

		LaunchParams( dim3 IBlocks, dim3 IThreads, cudaStream_t IStream) {
			Blocks = IBlocks;
			Threads = IThreads;
			Stream = IStream;
			SharedMem = 0;
		}

		dim3 Blocks;
		dim3 Threads;
		size_t SharedMem;
		cudaStream_t Stream;
	};

	// ---------------------------------------------------
	// ------- ------- Main Device Buffers ------- -------
	// ---------------------------------------------------

	struct DeviceArray {

		// -----------------------------------
		// ------- DEFAULT CONSTRUCTOR -------
		DeviceArray() {
			Dims = Dimensions( 0, 0, 0 );
			DataFormat = ArrayFormat::QS_ROW_MAJOR;
			Borrowed = false;
			Available = false;
			ArraySize = 0;
			Ptr = NULL;
			ID = -1;
		 }

		// ---------------------------------------------------
		// ------- CONSTRUCTOR WITH CUSTOM DATA FORMAT -------
		DeviceArray(Dimensions IDims,ArrayFormat Format,int IID) {
			size_t S_X = static_cast<size_t>(IDims.X);
			size_t S_Y = static_cast<size_t>(IDims.Y);
			size_t S_Z = static_cast<size_t>(IDims.Z);
			ArraySize = static_cast< size_t >( S_X * S_Y * S_Z * static_cast<size_t>(sizeof( cufftReal )) );
			cudaMalloc( ( void** ) &Ptr, ArraySize );
			Borrowed = false;
			DataFormat = Format;
			Available = false;
			Dims = IDims; 
			ID = IID;
		 }

		// -----------------------------------------------------
		// -------- CONSTRUCTOR WITH DEFAULT DATA FORMAT -------
		DeviceArray( Dimensions IDims, int IID ) {
			size_t S_X = static_cast<size_t>(IDims.X);
			size_t S_Y = static_cast<size_t>(IDims.Y);
			size_t S_Z = static_cast<size_t>(IDims.Z);
			ArraySize = static_cast< size_t >(S_X * S_Y * S_Z * static_cast<size_t>( sizeof( cufftReal )) );
			cudaMalloc( ( void** ) &Ptr, ArraySize );
			DataFormat = ArrayFormat::QS_ROW_MAJOR;
			Borrowed = false;
			Available = false;
			Dims = IDims;
			ID = IID;
		}

		// ---------------------------------
		// ------- SIMPLE DESTRUCTOR -------
		~DeviceArray() {
			if( Ptr != NULL ) { cudaFree( Ptr ); }
			Available = false;
			Borrowed = false;
			Ptr = NULL;
			ArraySize = 0;
			Dims = Dimensions( 0, 0, 0 );
		 }

		ArrayFormat DataFormat;	 // EITHER ROW_MAJOR OR COLUMN_MAJOR
		Dimensions Dims;		 // DIMENSIONS OF THE ALLOCATED PIECE OF MEMORY
		cufftReal *Ptr;			 // POINTER TO THE ALLOCATED MEMORY
		size_t ArraySize;		 // SIZE IN BYTES OF THE ALLOCATED AREA
		bool Available;			 // WHEATHER IT IS USED OR NOT. USED ARRAYS CANNOT BE ASSIGNED DURING "GETBUFFER"
		bool Borrowed;			 // ONLY METHOD CALLING "CREATE" HAS DESTROY RIGHTS
		int ID;				     // UNIQUE ID OF THE BUFFER
		
	};

	// -------------------------------------------------
	// ------- ------- Main Host Buffers ------- -------
	// -------------------------------------------------

	struct HostArray {
		
		// -----------------------------------
		// ------- DEFAULT CONSTRUCTOR -------
		HostArray() {
			Dims = Dimensions( 0, 0, 0 );
			DataFormat = ArrayFormat::QS_ROW_MAJOR;
			Available = true;
			Borrowed = false;
			ArraySize = 0;
			Ptr = NULL;
			ID = -1;
		}

		// ---------------------------------------------------
		// ------- CONSTRUCTOR WITH CUSTOM DATA FORMAT -------
		HostArray( Dimensions IDims ,ArrayFormat Format,int IID) {
			size_t S_X = static_cast<size_t>(IDims.X);
			size_t S_Y = static_cast<size_t>(IDims.Y);
			size_t S_Z = static_cast<size_t>(IDims.Z);
			ArraySize = static_cast<size_t>(S_X * S_Y * S_Z * static_cast<size_t>(sizeof(cufftReal)));
			cudaHostAlloc( ( void** ) &Ptr, ArraySize, cudaHostAllocDefault );
			DataFormat = Format;
			Available = false;
			Borrowed = false;
			Dims = IDims;
			ID = IID;

		}
		
		// -----------------------------------------------------
		// -------- CONSTRUCTOR WITH DEFAULT DATA FORMAT -------
		HostArray( Dimensions IDims, int IID) {
			size_t S_X = static_cast<size_t>(IDims.X);
			size_t S_Y = static_cast<size_t>(IDims.Y);
			size_t S_Z = static_cast<size_t>(IDims.Z);
			ArraySize = static_cast<size_t>(S_X * S_Y * S_Z * static_cast<size_t>(sizeof(cufftReal)));
			cudaHostAlloc( ( void** ) &Ptr, ArraySize, cudaHostAllocDefault );
			DataFormat = ArrayFormat::QS_ROW_MAJOR;
			Available = false;
			Borrowed = false;
			Dims = IDims;
			ID = IID;

		}

		// ---------------------------------
		// ------- SIMPLE DESTRUCTOR -------
		~HostArray() {
			if( Ptr != NULL ) { cudaFreeHost( Ptr ); }
			Dims = Dimensions( 0, 0, 0 );
			Available = false;
			Borrowed = false;
			ArraySize = 0;
			Ptr = NULL;	
		}

		ArrayFormat DataFormat;
		Dimensions Dims;
		cufftReal *Ptr;
		size_t ArraySize;
		bool Available;
		bool Borrowed;
		int ID;
	};

	struct IOArray {
		IOArray() {
			RealPtr = NULL;
			Dims = Dimensions( 1, 1, 1 );
			Complexity = ArrayComplexity::QS_REAL;
			Precision = ArrayPrecision::QS_SINGLE_PRECISION;
			Format = ArrayFormat::QS_ROW_MAJOR;
			
		}
		cufftReal *RealPtr;
		Dimensions Dims;
		ArrayComplexity Complexity;
		ArrayPrecision Precision;
		ArrayFormat Format;
	};

}