#include "QSwift.h"

// ----------------------------
// ------- C++ Includes -------
#include <iostream>
#include <string>

// -----------------------------
// ------ Matlab Includes ------
#include "mex.h"
#include "matrix.h"


// -----------------------
// ------- Globals -------
vector<IOArray> InputArray;
vector<IOArray> OutputArray;
ComputingType Type;

/************************************
 *@brief: Convert Input Matlab Arrays
 ************************************/
void MxArrayToIOArray(IOArray *QRay, const mxArray *MatArray) {

    const mwSize *Dimensions = mxGetDimensions( MatArray );
	const mwSize NumDims = mxGetNumberOfDimensions( MatArray );
	const mwSize NumElem = mxGetNumberOfElements( MatArray );
	mxClassID CID = mxGetClassID( MatArray );
	
	// ---------------------------------
	// ------ Extract Dimensions -------
	if(NumDims <= 2) {
		QRay->Dims.Y = static_cast<int>(Dimensions[0]);
		QRay->Dims.X = static_cast<int>(Dimensions[1]);
		QRay->Dims.Z = 1;
	}else {
		QRay->Dims.Y = static_cast<int>(Dimensions[0]);
		QRay->Dims.X = static_cast<int>(Dimensions[1]);
		QRay->Dims.Z = static_cast<int>(Dimensions[2]);
	}

	// ---------------------------------
	// ------- Prepare Data Array ------
	QRay->RealPtr = new float[NumElem];
	

	// --------------------------------------------------------------
	if (CID == mxDOUBLE_CLASS) { // ------- Double Precision --------
	// --------------------------------------------------------------
		QRay->Precision = ArrayPrecision::QS_DOUBLE_PRECISION;
		double* DPtr = mxGetPr(MatArray);
		mexWarnMsgTxt("Converting input to single precision! \n\0");
		for (auto i = 0; i < NumElem; i++) {
			QRay->RealPtr[i] = static_cast<float>(DPtr[i]);
		}
	// -----------------------------------------------------------------
	}else if(CID == mxSINGLE_CLASS) { // ------ Single Precision -------
	// -----------------------------------------------------------------

		QRay->Precision = ArrayPrecision::QS_SINGLE_PRECISION;
		float* SPtr = static_cast<float*>(mxGetData( MatArray ));
		memcpy( QRay->RealPtr, SPtr, NumElem * sizeof( float ) );
	}else {
		mexErrMsgTxt("Invalid Array Class. Only Single/Double Precision is Supported! \n\0");
	}
	
}

/******************************************************
 *@brief Parses Input Matlab Array for a single Integer  
 ******************************************************/
int ParseMxArrayInt( const mxArray *MatArray ) {
	int OutputParam = 0;
	mxClassID CID = mxGetClassID( MatArray );
	if( CID == mxSINGLE_CLASS ) {
		cufftReal *Temp = static_cast<cufftReal*>( mxGetData( MatArray ) );
		OutputParam = static_cast<int>(Temp[0]);
	} else if( CID == mxDOUBLE_CLASS ) {
		cufftDoubleReal *Temp = static_cast< cufftDoubleReal* >( mxGetData( MatArray ) );
		OutputParam = static_cast< int >( Temp[0] );
	} else if( CID == mxINT32_CLASS){
		int *Temp = static_cast< int* >( mxGetData( MatArray ) );
		OutputParam = Temp[0];
	} else {
		string message = string( "Unsupported class for precision: " ) + string( mxGetClassName( MatArray ) ) + string( "\n\0" );
		mexErrMsgTxt( message.c_str() );
	}

	return OutputParam;
}

/*****************************************************
 *@brief: Parses Input Matlab Array for a single Float 
 ****************************************************/
float ParseMxArrayFloat( const mxArray *MatArray ) {

	float OutputParam = 0.0f;
	mxClassID CID = mxGetClassID( MatArray );
	if( CID == mxSINGLE_CLASS ) {
		float *Temp = static_cast<float*>( mxGetData( MatArray ) );
		OutputParam = Temp[0];
	} else if( CID == mxDOUBLE_CLASS ) {
		double *Temp = static_cast< double* >( mxGetData( MatArray ) );
		OutputParam = static_cast< float >( Temp[0] );
	}else if( CID  == mxINT32_CLASS){
		int *Temp = static_cast< int* >( mxGetData( MatArray ) );
		OutputParam = static_cast< float >( Temp[0] );
	} else {
		string message = string( "Unsupported class for precision: " ) + string( mxGetClassName( MatArray ) ) + string( "\n\0" );
		mexErrMsgTxt( message.c_str() );
	}
	return OutputParam;
}

/**************************************************
 *@brief: Parses All Required Computing Parameters 
 *************************************************/
bool ParseInput(QSwift *QS,const mxArray *prhs[],int NumParams){

	Type = ComputingType::QS_METHOD_UNKNOWN;

	// ------------------------------------------------
	// ------- Loop across all Input Parameters -------
	for(int i = 0;i < NumParams - 1 ;i++){
		const mxArray *InputParam = prhs[i];
		if( mxGetClassID( InputParam ) != mxCHAR_CLASS ){
			string message = "Error parsing: " + to_string( i ) + "th parameter! \n\0";
			mexErrMsgTxt(message.c_str());
			return false;
		}

		char *UserInput = mxArrayToString( InputParam );
		string ToCompare = string( UserInput );

		// --- --- --- --- --- --- --- --- ---
		if(ToCompare.compare("IData1") == 0){
			IOArray Input1;
			const mxArray *MatData = prhs[i + 1];
			MxArrayToIOArray( &Input1, MatData );
			InputArray.insert( InputArray.begin(), Input1 );
		}

		// --- --- --- --- --- --- --- --- ---
		if(ToCompare.compare("InitialGuess") == 0) {
			IOArray Input3;
			const mxArray *MatData = prhs[i + 1];
			MxArrayToIOArray( &Input3, MatData );
			InputArray.insert( InputArray.begin() + 2, Input3 );
		}

		// --- --- --- --- --- --- --- --- ---
		if(ToCompare.compare("IData2") == 0){
			IOArray Input2;
			const mxArray *MatData = prhs[i + 1];
			MxArrayToIOArray( &Input2,MatData );
			InputArray.insert( InputArray.begin() + 1, Input2 );
		}

		// --- --- --- --- --- --- --- --- ---
		if(ToCompare.compare("Precision") == 0){
			const mxArray *MatData = prhs[i + 1];
			float Precision = ParseMxArrayFloat( MatData );
			QS->SetParameter( Precision, QS_PRECISION );
		}

		// --- --- --- --- --- --- --- --- ---
		if(ToCompare.compare("Omega") == 0) {
			const mxArray *MatData = prhs[i + 1];
			float Omega = ParseMxArrayFloat( MatData );
			QS->SetParameter( Omega, QS_GREEK_OMEGA );
		}

		// --- --- --- --- --- --- --- --- ---
		if(ToCompare.compare("NumIter") == 0){
			const mxArray *MatData = prhs[i + 1];
			int NumIter = ParseMxArrayInt( MatData );
			QS->SetParameter( NumIter, QS_NUM_ITER );
		}

		// --- --- --- --- --- --- --- --- ---
		if(ToCompare.compare("Type") == 0){
			const mxArray *MatData = prhs[i+1];
			if( mxGetClassID( MatData ) != mxCHAR_CLASS ) {
				string message = "Error parsing: " + to_string( i ) + "th parameter! \n\0";
				mexErrMsgTxt( message.c_str() );
				return false;
			}
			char *SelectedMethod= mxArrayToString( MatData );
			string CompareWith = string( SelectedMethod );
			
			if(CompareWith.compare( "MatrixMatrixMul") == 0){
				Type = ComputingType::QS_METHOD_MATRIXMUL;
			}else if( CompareWith.compare( "GaussElim") == 0){
				Type = ComputingType::QS_METHOD_GAUSSELIM;
			}else if( CompareWith.compare( "MatrixInverse") == 0){
				Type = ComputingType::QS_METHOD_INVERSION;
			}else if( CompareWith.compare( "Eigenvalues") == 0){
				Type = ComputingType::QS_METHOD_EIGENVALUES;
			}else if( CompareWith.compare( "GramSchmidt") == 0){
				Type = ComputingType::QS_METHOD_GRAM_SCHMIDT;
			} else if( CompareWith.compare( "HauseHolder" ) == 0 ) {
				Type = ComputingType::QS_METHOD_HAUSEHOLDER;
			} else if( CompareWith.compare( "Jacobi" ) == 0 ) {
				Type = ComputingType::QS_METHOD_JACOBIITER;
			} else if( CompareWith.compare( "Sor" ) == 0) {
				Type = ComputingType::QS_METHOD_SOR;
			}else{
				Type = ComputingType::QS_METHOD_UNKNOWN;
				string message = "Unknown Computing Type!: " + CompareWith;
				mexErrMsgTxt( message.c_str() );
				return false;
			}
			
		}

		// --- ----
		i++; // ---
		// --- ----
	}
	return true;
	
}

/*****************************************************
 * @brief: Matlab Entry Function
 * prhs = Pointer to The Right Hand Side  ==> INPUT
 * plhs = Pointer to The Left  Hand Side  ==> OUTPUT
 *****************************************************/ 
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] ) {
	
	/************************************
	 * Possible Combinations are:
	 * IData1 + Matrix / Vector
	 * IData2 + Matrix / Vector
	 * InitialGuess / vector
	 * Precision - Scalar
	 * Omega - Scalar
	 * NumIter - Scalar
	 * Type - String
	 *		--> "MatrixMatrixMul"
	 *		--> "GaussElim"
	 *		--> "MatrixInverse"
	 *		--> "Eigenvalues"
	 *		--> "GramSchmidt"
	 *		--> "HauseHolder"
	 *		-->	"Jacobi"
	 *		--> "Sor"
	 ************************************/
	if( nrhs < 4 ) { mexErrMsgTxt( "At least 4 Parameters are required!\n\0" ); }
	
	// ----------------------------------------------
	// ------- Initialize Acceleration Library ------
	QSwift *IrisCores = new QSwift();

	// ----------------------------------
	// ------- Parse Matlab Input -------
	ParseInput( IrisCores, prhs, nrhs );

	// EXTRACT TO GET CORRECT DIMENIONS
	IOArray InputMatrix = InputArray[0];

	// -------------------------------------
	// ------- Prepare Output Arrays -------
	switch(Type ){
		// ------------------------------------------------------------
		case ComputingType::QS_METHOD_EIGENVALUES:{ // ------ 1x VECTOR
		// ------------------------------------------------------------
			IOArray Output_1;
			size_t VectorDims[] = { static_cast<size_t>( 1 ), static_cast<size_t>( InputMatrix.Dims.X ) };
			const mxArray *OutputVector = plhs[0] = mxCreateNumericArray( 2, VectorDims, mxSINGLE_CLASS, mxREAL );
			Output_1.RealPtr = static_cast< float* >( mxGetData( OutputVector ) );
			OutputArray.push_back( Output_1 );
			break;
			}
		// -----------------------------------------------------------
		case ComputingType::QS_METHOD_GAUSSELIM:{ // ------- 1x MATRIX
		// -----------------------------------------------------------
			IOArray Output_1;
			size_t MatrixDims[] = { static_cast<size_t>( InputMatrix.Dims.Y ), static_cast<size_t>( InputMatrix.Dims.X ) };
			const mxArray *OutputMatrix = plhs[0] = mxCreateNumericArray( 2, MatrixDims, mxSINGLE_CLASS, mxREAL );
			Output_1.RealPtr = static_cast<float*>( mxGetData( OutputMatrix ) );
			OutputArray.push_back( Output_1 );
			break;
			}
		// ---------------------------------------------------------------
		case ComputingType::QS_METHOD_GRAM_SCHMIDT:{ // ------- 2 x MATRIX
		// ---------------------------------------------------------------
			IOArray Output_1;
			IOArray Output_2;
			size_t MatrixDims[] = { static_cast<size_t>( InputMatrix.Dims.Y ), static_cast<size_t>( InputMatrix.Dims.X ) };
			const mxArray *OutputMatrix1 = plhs[0] = mxCreateNumericArray( 2, MatrixDims, mxSINGLE_CLASS, mxREAL );
			const mxArray *OutputMatrix2 = plhs[1] = mxCreateNumericArray( 2, MatrixDims, mxSINGLE_CLASS, mxREAL );
			Output_1.RealPtr = static_cast<float*>( mxGetData( OutputMatrix1 ) );
			Output_2.RealPtr = static_cast<float*>( mxGetData( OutputMatrix2 ) );
			OutputArray.push_back( Output_1 );
			OutputArray.push_back( Output_2 );
			break;
			}
		// -------------------------------------------------------------
		case ComputingType::QS_METHOD_HAUSEHOLDER:{ // ------- 1X MATRIX
		// -------------------------------------------------------------
			IOArray Output_1;
			size_t MatrixDims[] = { static_cast<size_t>( InputMatrix.Dims.Y ), static_cast<size_t>( InputMatrix.Dims.X ) };
			const mxArray *OutputMatrix = plhs[0] = mxCreateNumericArray( 2, MatrixDims, mxSINGLE_CLASS, mxREAL );
			Output_1.RealPtr = static_cast<float*>( mxGetData( OutputMatrix ) );
			OutputArray.push_back( Output_1 );
			break;
			}
		// -----------------------------------------------------------
		case ComputingType::QS_METHOD_MATRIXMUL:{ // ------- 1X MATRIX
		// -----------------------------------------------------------
			IOArray Output_1;
			size_t MatrixDims[] = { static_cast<size_t>( InputMatrix.Dims.Y ), static_cast<size_t>( InputMatrix.Dims.X ) };
			const mxArray *OutputMatrix = plhs[0] = mxCreateNumericArray( 2, MatrixDims, mxSINGLE_CLASS, mxREAL );
			Output_1.RealPtr = static_cast<float*>( mxGetData( OutputMatrix ) );
			OutputArray.push_back( Output_1 );
			break;
			}
		// -----------------------------------------------------------
		case ComputingType::QS_METHOD_INVERSION:{ // ------- 1X MATRIX
		// -----------------------------------------------------------
			IOArray Output_1;
			size_t MatrixDims[] = { static_cast<size_t>( InputMatrix.Dims.Y ), static_cast<size_t>( InputMatrix.Dims.X ) };
			const mxArray *OutputMatrix = plhs[0] = mxCreateNumericArray( 2, MatrixDims, mxSINGLE_CLASS, mxREAL );
			Output_1.RealPtr = static_cast<float*>( mxGetData( OutputMatrix ) );
			OutputArray.push_back( Output_1 );
			break;
			}
		// -------------------------------------------------------------
		case ComputingType::QS_METHOD_JACOBIITER: { // ------- 1X VECTOR
		// -------------------------------------------------------------
			IOArray Output_1;
			size_t VectorDims[] = {  static_cast<size_t>( InputMatrix.Dims.Y ), static_cast<size_t>( 1 )};
			const mxArray *OutputVector = plhs[0] = mxCreateNumericArray( 2, VectorDims, mxSINGLE_CLASS, mxREAL );
			Output_1.RealPtr = static_cast< float* >( mxGetData( OutputVector ) );
			OutputArray.push_back( Output_1 );
			break;
		}
		case ComputingType::QS_METHOD_SOR: {
			// VECTOR
			IOArray Output_1;
			size_t VectorDims[] = { static_cast< size_t >( InputMatrix.Dims.Y ),static_cast< size_t >( 1 ) };
			const mxArray *OutputVector = plhs[0] = mxCreateNumericArray( 2, VectorDims, mxSINGLE_CLASS, mxREAL );
			Output_1.RealPtr = static_cast< float* >( mxGetData( OutputVector ) );
			OutputArray.push_back( Output_1 );
			break;
		}
		default:{	
			string message = string( "Error processing. Computation Type not specified! \n\0" );
			mexErrMsgTxt( message.c_str() );
			break;
		}
	}


	// 
	IrisCores->SetInput( InputArray );
	IrisCores->SetOutput( OutputArray );
	IrisCores->SetComputingType( Type );
	bool Succeeded = IrisCores->Init( );
	if( Succeeded ) { IrisCores->Compute(); } else{
		vector<string> EBuffer = IrisCores->GetErrorBuffer();
		for( int i = 0; i<EBuffer.size(); i++ ) {
			mexWarnMsgTxt( EBuffer[i].c_str() );
		}
		mexErrMsgTxt( "SwiftQR Init Failure! \n\0" );
	}

	// -------------------------------------------------------
	// ------- Get All Stored Messages And Display Them ------
	vector<string> EBuffer = IrisCores->GetErrorBuffer();
	vector<string> MBuffer = IrisCores->GetMessageBuffer();
	if( MBuffer.size() > 0 ) {
		for( int i = 0; i < MBuffer.size(); i++ ) {
			mexPrintf( MBuffer[i].c_str() );
		}
	}
	if(EBuffer.size() > 0){
		for(int i = 0; i < EBuffer.size(); i++){
			mexWarnMsgTxt( EBuffer[i].c_str() );
		}
		string message = "QSwift encountered a total of: " + to_string( EBuffer.size() ) + string( " Errors \n\0" );
		mexErrMsgTxt( message.c_str() );
	}

	// ----------------------------
	// ------- CLEAN MEMORY -------
	for(int i = 0;i<InputArray.size();i++) {delete [] (InputArray[i].RealPtr);}
	InputArray.clear();
	delete (IrisCores);
}
